# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

"""General editor panel utilities."""

# Standard library imports
import bisect
import uuid

# Third-party imports
from intervaltree import IntervalTree

# --------------------- Code Folding Panel ------------------------------------


class FoldingRegion:
    """Internal representation of a code folding region."""

    def __init__(self, text, fold_range):
        self.text = text
        self.fold_range = fold_range
        self.id = str(uuid.uuid4())
        self.index = None
        self.nesting = 0
        self.children = []
        self.status = False
        self.parent = None

    def delete(self):
        for child in self.children:
            child.parent = None

        self.children = []
        if self.parent is not None:
            self.parent.remove_node(self)
            self.parent = None

    def add_node(self, node):
        node.parent = self
        node.nesting = self.nesting + 1
        children_ranges = [c.fold_range[0] for c in self.children]
        node_range = node.fold_range[0]
        new_index = bisect.bisect_left(children_ranges, node_range)
        node.index = new_index
        for child in self.children[new_index:]:
            child.index += 1
        self.children.insert(new_index, node)

    def remove_node(self, node):
        for child in self.children[node.index + 1:]:
            child.index -= 1

        try:
            self.children.pop(node.index)
        except IndexError:
            pass
        for idx, next_idx in zip(self.children, self.children[1:]):
            assert idx.index < next_idx.index

    def clone_node(self, node):
        self.id = node.id
        self.index = node.index
        self.nesting = node.nesting
        self.parent = node.parent
        self.children = node.children
        self.status = node.status

        for child in self.children:
            child.parent = self

        if self.parent is not None:
            self.parent.replace_node(self.index, self)

    def replace_node(self, index, node):
        self.children[index] = node
        node.parent = self

    def __repr__(self):
        return str(self)

    def __str__(self):
        return '({0}, {1}, {2}, {3})'.format(
            self.fold_range, self.text, self.id, self.status)


class FoldingStatus(dict):
    """
    Code folding status storage.

    This dictionary subclass is used to update and get the status of a
    folding region without having to deal with the internal representation.
    """

    def values(self):
        values = dict.values(self)
        return [x.status for x in values]

    def __getitem__(self, key):
        value = dict.__getitem__(self, key)
        return value.status

    def __setitem__(self, key, value):
        if isinstance(value, FoldingRegion):
            dict.__setitem__(self, key, value)
        else:
            region = dict.__getitem__(self, key)
            region.status = value


def merge_interval(parent, node):
    """Build code folding tree representation from interval tree."""
    match = False
    start, __ = node.fold_range
    while parent.parent is not None and not match:
        __, parent_end = parent.fold_range
        if parent_end <= start:
            parent = parent.parent
        else:
            match = True

    if node.parent is not None:
        node.parent.remove_node(node)
        node.parent = None

    parent.add_node(node)
    return node


def merge_folding(ranges, current_tree, root):
    """Compare previous and current code folding tree information."""
    # Leave this import here to avoid importing Numpy (which is used by
    # textdistance) too early at startup.
    import textdistance

    folding_ranges = []
    for starting_line, ending_line, text in ranges:
        if ending_line > starting_line:
            starting_line += 1
            ending_line += 1
            folding_repr = FoldingRegion(text, (starting_line, ending_line))
            folding_ranges.append((starting_line, ending_line, folding_repr))

    tree = IntervalTree.from_tuples(folding_ranges)
    changes = tree - current_tree
    deleted = current_tree - tree
    adding_folding = len(changes) > len(deleted)

    deleted_iter = iter(sorted(deleted))
    changes_iter = iter(sorted(changes))
    deleted_entry = next(deleted_iter, None)
    changed_entry = next(changes_iter, None)
    non_merged = 0

    while deleted_entry is not None and changed_entry is not None:
        deleted_entry_i = deleted_entry.data
        changed_entry_i = changed_entry.data
        dist = textdistance.hamming.normalized_similarity(
            deleted_entry_i.text, changed_entry_i.text)

        if dist >= 0.80:
            # Copy folding status
            changed_entry_i.clone_node(deleted_entry_i)
            deleted_entry = next(deleted_iter, None)
            changed_entry = next(changes_iter, None)
        else:
            if adding_folding:
                # New symbol added
                non_merged += 1
                changed_entry = next(changes_iter, None)
            else:
                # Symbol removed
                deleted_entry_i.delete()
                non_merged += 1
                deleted_entry = next(deleted_iter, None)

    if deleted_entry is not None:
        while deleted_entry is not None:
            # Symbol removed
            deleted_entry_i = deleted_entry.data
            deleted_entry_i.delete()
            non_merged += 1
            deleted_entry = next(deleted_iter, None)

    if changed_entry is not None:
        while changed_entry is not None:
            non_merged += 1
            changed_entry = next(changes_iter, None)

    if non_merged > 0:
        tree_copy = IntervalTree(tree)
        tree_copy.merge_overlaps(
            data_reducer=merge_interval,
            data_initializer=root)
    return tree, root


def collect_folding_regions(root):
    queue = [(x, 0, -1) for x in root.children]
    folding_status = FoldingStatus({})
    folding_regions = {}
    folding_nesting = {}
    folding_levels = {}
    while queue != []:
        node, folding_level, folding_nest = queue.pop(0)
        start, end = node.fold_range
        folding_regions[start] = end
        folding_levels[start] = folding_level
        folding_nesting[start] = folding_nest
        folding_status[start] = node
        queue = [(x, folding_level + 1, start) for x in node.children] + queue
    return folding_regions, folding_nesting, folding_levels, folding_status



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

"""
Cookiecutter utilities.
"""

import json
import os

from cookiecutter.main import cookiecutter


def generate_cookiecutter_project(cookiecutter_path, output_path,
                                  extra_content=None):
    """
    Generate a cookicutter project programmatically.
    """
    status = True
    try:
        result = cookiecutter(
            cookiecutter_path,
            output_dir=output_path,
            overwrite_if_exists=True,
            extra_context=extra_content,
            no_input=True,
        )
    except Exception as err:
        result = err
        status = False

    return status, result


def load_cookiecutter_project(project_path):
    """
    Load a cookicutter options and pre-hook script.
    """
    options = None
    pre_gen_code = None
    cookiepath = os.path.join(project_path, "cookiecutter.json")
    pre_gen_path = os.path.join(project_path, "hooks", "pre_gen_project.py")

    if os.path.isdir(project_path):
        if os.path.isfile(cookiepath):
            with open(cookiepath, 'r') as fh:
                options = json.loads(fh.read())

        if os.path.isfile(pre_gen_path):
            with open(pre_gen_path, 'r') as fh:
                pre_gen_code = fh.read()

    return options, pre_gen_code



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

"""
Status bar container.
"""

# Third-party imports
from qtpy.QtCore import Signal

# Local imports
from spyder.api.config.decorators import on_conf_change
from spyder.api.widgets.main_container import PluginMainContainer
from spyder.plugins.statusbar.widgets.status import (
    ClockStatus, CPUStatus, MemoryStatus
)


class StatusBarContainer(PluginMainContainer):

    sig_show_status_bar_requested = Signal(bool)
    """
    This signal is emmitted when the user wants to show/hide the
    status bar.
    """

    def setup(self):
        # Basic status widgets
        self.mem_status = MemoryStatus(parent=self)
        self.cpu_status = CPUStatus(parent=self)
        self.clock_status = ClockStatus(parent=self)

    @on_conf_change(option='memory_usage/enable')
    def enable_mem_status(self, value):
        self.mem_status.setVisible(value)

    @on_conf_change(option='memory_usage/timeout')
    def set_mem_interval(self, value):
        self.mem_status.set_interval(value)

    @on_conf_change(option='cpu_usage/enable')
    def enable_cpu_status(self, value):
        self.cpu_status.setVisible(value)

    @on_conf_change(option='cpu_usage/timeout')
    def set_cpu_interval(self, value):
        self.cpu_status.set_interval(value)

    @on_conf_change(option='clock/enable')
    def enable_clock_status(self, value):
        self.clock_status.setVisible(value)

    @on_conf_change(option='clock/timeout')
    def set_clock_interval(self, value):
        self.clock_status.set_interval(value)

    @on_conf_change(option='show_status_bar')
    def show_status_bar(self, value):
        self.sig_show_status_bar_requested.emit(value)

    def update_actions(self):
        pass



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright © Spyder Project Contributors
#
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)
# -----------------------------------------------------------------------------

"""
Cookiecutter widget.
"""

import sys
import tempfile
from collections import OrderedDict

from jinja2 import Template
from qtpy import QtCore
from qtpy import QtWidgets


class Namespace:
    """
    Namespace to provide a holder for attributes when rendering a template.
    """

    def __init__(self, **kwargs):
        for key, value in kwargs.items():
            setattr(self, key, value)


class CookiecutterDialog(QtWidgets.QDialog):
    """
    QDialog to display cookiecutter.json options.

    cookiecutter_settings: dict
        A cookiecutter.json settings content.
    pre_gen_code: str
        The code of the pregeneration script.
    """

    sig_validated = QtCore.Signal(int, str)
    """
    This signal is emitted after validation has been executed.

    It provides the process exit code and the output captured.
    """

    def __init__(self, parent, cookiecutter_settings=None, pre_gen_code=None):
        super().__init__(parent)

        self._widget = CookiecutterWidget(
            self, cookiecutter_settings,
            pre_gen_code
        )
        self._info_label = QtWidgets.QLabel()
        self._validate_button = QtWidgets.QPushButton("Validate")

        layout = QtWidgets.QVBoxLayout()
        layout.addWidget(self._widget)
        layout.addWidget(self._info_label)
        layout.addWidget(self._validate_button)
        self.setLayout(layout)

        # Signals
        self._validate_button.clicked.connect(self.validate)
        self._widget.sig_validated.connect(self._set_message)
        self._widget.sig_validated.connect(self.sig_validated)

    def _set_message(self, exit_code, message):
        if exit_code != 0:
            self._info_label.setText(message)

    def setup(self, cookiecutter_settings):
        """
        Setup the widget using options.
        """
        self._widget.setup(cookiecutter_settings)

    def set_pre_gen_code(self, pre_gen_code):
        """
        Set the cookiecutter pregeneration code.
        """
        self._widget.set_pre_gen_code(pre_gen_code)

    def validate(self):
        """
        Run, pre generation script and provide information on finished.
        """
        self._widget.validate()

    def get_values(self):
        """
        Return all entered and generated values.
        """
        return self._widget.get_values()


class CookiecutterWidget(QtWidgets.QWidget):
    """
    QWidget to display cookiecutter.json options.

    cookiecutter_settings: dict
        A cookiecutter.json settings content.
    pre_gen_code: str
        The code of the pregeneration script.
    """

    sig_validated = QtCore.Signal(int, str)
    """
    This signal is emitted after validation has been executed.

    It provides the process exit code and the output captured.
    """

    def __init__(self, parent, cookiecutter_settings=None, pre_gen_code=None):
        super().__init__(parent)

        # Attributes
        self._parent = parent
        self._cookiecutter_settings = cookiecutter_settings
        self._pre_gen_code = pre_gen_code
        self._widgets = OrderedDict()
        self._defined_settings = OrderedDict()
        self._rendered_settings = OrderedDict()
        self._process = None
        self._tempfile = tempfile.mkstemp(suffix=".py")[-1]

        # Cookiecutter special variables
        self._extensions = None
        self._copy_without_render = None
        self._new_lines = None
        self._private_vars = None
        self._rendered_private_var = None

        # Layout
        self._form_layout = QtWidgets.QFormLayout()
        self._form_layout.setFieldGrowthPolicy(
            self._form_layout.AllNonFixedFieldsGrow)
        self.setLayout(self._form_layout)

    # --- Helpers
    # ------------------------------------------------------------------------
    def _check_jinja_options(self):
        """
        Check which values are Jinja2 expressions.
        """
        if self._cookiecutter_settings:
            # https://cookiecutter.readthedocs.io/en/latest/advanced/template_extensions.html
            self._extensions = self._cookiecutter_settings.pop("_extensions",
                                                               [])

            # https://cookiecutter.readthedocs.io/en/latest/advanced/copy_without_render.html
            self._copy_without_render = self._cookiecutter_settings.pop(
                "_copy_without_render", [])

            # https://cookiecutter.readthedocs.io/en/latest/advanced/new_line_characters.html
            self._new_lines = self._cookiecutter_settings.pop("_new_lines", "")

            for setting, value in self._cookiecutter_settings.items():
                # Treat everything like a list for convenience
                if isinstance(value, dict):
                    # https://cookiecutter.readthedocs.io/en/latest/advanced/dict_variables.html
                    list_values = list(value.keys())
                elif not isinstance(value, list):
                    list_values = [value]
                else:
                    list_values = value

                are_rendered_values = []
                if list_values and value:
                    for list_value in list_values:
                        template = Template(list_value)
                        rendered_value = template.render(
                            cookiecutter=Namespace(
                                **self._cookiecutter_settings))

                        are_rendered_values.append(
                            list_value != rendered_value)

                if any(are_rendered_values):
                    self._rendered_settings[setting] = value
                else:
                    self._defined_settings[setting] = value

    def _is_jinja(self, setting):
        """
        Check if option contains jinja2 code.
        """
        return setting in self._rendered_settings

    def _parse_bool_text(self, text):
        """
        Convert a text value into a boolean.
        """
        value = None
        if text.lower() in ["n", "no", "false"]:
            value = False
        elif text.lower() in ["y", "yes", "true"]:
            value = True

        return value

    def _create_textbox(self, setting, label, default=None):
        """
        Create a textbox field.
        """
        if default is not None and len(default) > 30:
            box = QtWidgets.QTextEdit(parent=self)
            box.setText = box.setPlainText
            box.text = box.toPlainText
        else:
            box = QtWidgets.QLineEdit(parent=self)

        box.setting = setting
        if default is not None:
            box.setText(default)
            box.textChanged.connect(lambda x=None: self.render())

        box.get_value = box.text
        box.set_value = lambda text: box.setText(text)

        return box

    def _create_checkbox(self, setting, label, default=None):
        """
        Create a checkbox field.
        """
        box = QtWidgets.QCheckBox(parent=self)
        box.setting = setting
        if default is not None:
            new_default = self._parse_bool_text(default)
            box.setChecked(new_default)

        def _get_value():
            bool_to_values = {
                self._parse_bool_text(default): default,
                not self._parse_bool_text(default): "other-value-" + default
            }
            return bool_to_values[box.isChecked()]

        box.get_value = _get_value

        return box

    def _create_combobox(self, setting, label, choices, default=None):
        """
        Create a combobox field.
        """
        box = QtWidgets.QComboBox(parent=self)
        if isinstance(choices, dict):
            temp = OrderedDict()
            for choice, choice_value in choices.items():
                box.addItem(choice, {choice: choice_value})
        else:
            for choice in choices:
                box.addItem(choice, choice)

        box.setting = setting
        box.get_value = box.currentData

        return box

    def _create_field(self, setting, value):
        """
        Create a form field.
        """
        label = " ".join(setting.split("_")).capitalize()
        if isinstance(value, (list, dict)):
            # https://cookiecutter.readthedocs.io/en/latest/advanced/choice_variables.html
            widget = self._create_combobox(setting, label, value)
        elif isinstance(value, str):
            if value.lower() in ["y", "yes", "true", "n", "no", "false"]:
                widget = self._create_checkbox(setting, label, default=value)
            else:
                default = None if self._is_jinja(setting) else value
                widget = self._create_textbox(setting, label, default=default)
        else:
            raise Exception(
                "Cookiecutter option '{}'cannot be processed".format(setting))

        self._widgets[setting] = (label, widget)

        return label, widget

    def _on_process_finished(self):
        """
        Process output of valiation script.
        """
        if self._process is not None:
            out = bytes(self._process.readAllStandardOutput()).decode()
            error = bytes(self._process.readAllStandardError()).decode()
            message = ""
            if out:
                message += out

            if error:
                message += error

            message = message.replace("\r\n", " ")
            message = message.replace("\n", " ")
            self.sig_validated.emit(self._process.exitCode(), message)

    # --- API
    # ------------------------------------------------------------------------
    def setup(self, cookiecutter_settings):
        """
        Setup the widget using options.
        """
        self._cookiecutter_settings = cookiecutter_settings
        self._check_jinja_options()

        for setting, value in self._cookiecutter_settings.items():
            if not setting.startswith(("__", "_")):
                label, widget = self._create_field(setting, value)
                self._form_layout.addRow(label, widget)

        self.render()

    def set_pre_gen_code(self, pre_gen_code):
        """
        Set the cookiecutter pregeneration code.
        """
        self._pre_gen_code = pre_gen_code

    def render(self):
        """
        Render text that contains Jinja2 expressions and set their values.
        """
        cookiecutter_settings = self.get_values()
        for setting, value in self._rendered_settings.items():
            if not setting.startswith(("__", "_")):
                template = Template(value)
                val = template.render(
                    cookiecutter=Namespace(**cookiecutter_settings))
                __, widget = self._widgets[setting]
                widget.set_value(val)

    def get_values(self):
        """
        Return all entered and generated values.
        """
        cookiecutter_settings = cs = OrderedDict()
        if self._cookiecutter_settings:
            for setting, value in self._cookiecutter_settings.items():
                if setting.startswith(("__", "_")):
                    cookiecutter_settings[setting] = value
                else:
                    __, widget = self._widgets[setting]
                    cookiecutter_settings[setting] = widget.get_value()

        # Cookiecutter special variables
        cookiecutter_settings["_extensions"] = self._extensions
        cookiecutter_settings["_copy_without_render"] = (
            self._copy_without_render)
        cookiecutter_settings["_new_lines"] = self._new_lines

        return cookiecutter_settings

    def validate(self):
        """
        Run, pre generation script and provide information on finished.
        """
        if self._pre_gen_code is not None:
            cookiecutter_settings = self.get_values()
            template = Template(self._pre_gen_code)
            val = template.render(
                cookiecutter=Namespace(**cookiecutter_settings))

            with open(self._tempfile, "w") as fh:
                fh.write(val)

            if self._process is not None:
                self._process.close()
                self._process.waitForFinished(1000)

            self._process = QtCore.QProcess(self)
            self._process.setProgram(sys.executable)
            self._process.setArguments([self._tempfile])
            self._process.finished.connect(self._on_process_finished)
            self._process.start()


if __name__ == "__main__":
    from spyder.utils.qthelpers import qapplication

    app = qapplication()
    dlg = CookiecutterDialog(parent=None)
    dlg.setup(
        {
            "list_option": ["1", "2", "3"],
            "checkbox_option": "y",
            "checkbox_option_2": "false",
            "fixed_option": "goanpeca",
            "rendered_option": "{{ cookiecutter.fixed_option|upper }}",
            "dict_option": {
                "png": {
                    "name": "Portable Network Graphic",
                    "library": "libpng",
                    "apps": [
                        "GIMP"
                    ]
                },
                "bmp": {
                    "name": "Bitmap",
                    "library": "libbmp",
                    "apps": [
                        "Paint",
                        "GIMP"
                    ]
                }
            },
            "_private": "{{ cookiecutter.fixed_option }}",
            "__private_rendered": "{{ cookiecutter.fixed_option }}",
        }
    )
    dlg.set_pre_gen_code('''
import sys
print("HELP!")  # spyder: test-skip
sys.exit(10)''')
    dlg.show()
    sys.exit(app.exec_())



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

"""
Spyder preferences plugin.

This plugin is in charge of managing preference pages and tabs for all plugins
in Spyder, both internal and external.
"""

# Standard library imports
import os
import logging
from typing import Union
from packaging.version import Version
from pkg_resources import parse_version

# Third-party imports
from qtpy.QtGui import QIcon
from qtpy.QtCore import Slot
from qtpy.QtWidgets import QMessageBox

# Local imports
from spyder.api.plugins import Plugins, SpyderPluginV2, SpyderPlugin
from spyder.api.plugin_registration.decorators import (
    on_plugin_available, on_plugin_teardown)
from spyder.config.base import _
from spyder.config.main import CONF_VERSION
from spyder.config.user import NoDefault
from spyder.plugins.mainmenu.api import ApplicationMenus, ToolsMenuSections
from spyder.plugins.preferences.widgets.container import (
    PreferencesActions, PreferencesContainer)
from spyder.plugins.pythonpath.api import PythonpathActions
from spyder.plugins.toolbar.api import ApplicationToolbars, MainToolbarSections

logger = logging.getLogger(__name__)

BaseType = Union[int, float, bool, complex, str, bytes]
IterableType = Union[list, tuple]
BasicType = Union[BaseType, IterableType]


class Preferences(SpyderPluginV2):
    """
    Spyder preferences plugin.

    This class manages all the preference pages and tabs for all internal
    and external plugins, as well enabling other plugins to add configurations
    to other sections.
    """

    NAME = 'preferences'
    CONF_SECTION = 'preferences'
    OPTIONAL = [Plugins.MainMenu, Plugins.Toolbar]
    CONF_FILE = False
    CONTAINER_CLASS = PreferencesContainer
    CAN_BE_DISABLED = False

    NEW_API = 'new'
    OLD_API = 'old'

    def __init__(self, parent, configuration=None):
        super().__init__(parent, configuration)
        self.config_pages = {}
        self.config_tabs = {}

    def register_plugin_preferences(
            self, plugin: Union[SpyderPluginV2, SpyderPlugin]) -> None:
        if (hasattr(plugin, 'CONF_WIDGET_CLASS') and
                plugin.CONF_WIDGET_CLASS is not None):
            # New API
            Widget = plugin.CONF_WIDGET_CLASS

            self.config_pages[plugin.NAME] = (self.NEW_API, Widget, plugin)

            plugin_conf_version = plugin.CONF_VERSION or CONF_VERSION
            plugin_conf_version = parse_version(plugin_conf_version)

            # Check if the plugin adds new configuration options to other
            # sections
            if plugin.ADDITIONAL_CONF_OPTIONS is not None:
                for conf_section in plugin.ADDITIONAL_CONF_OPTIONS:
                    conf_keys = plugin.ADDITIONAL_CONF_OPTIONS[conf_section]
                    for conf_key in conf_keys:
                        new_value = conf_keys[conf_key]
                        self.check_version_and_merge(
                            conf_section, conf_key, new_value,
                            plugin_conf_version, plugin)

            # Check if the plugin declares any additional configuration tabs
            if plugin.ADDITIONAL_CONF_TABS is not None:
                for plugin_name in plugin.ADDITIONAL_CONF_TABS:
                    tabs_to_add = plugin.ADDITIONAL_CONF_TABS[plugin_name]
                    plugin_tabs = self.config_tabs.get(plugin_name, [])
                    plugin_tabs += tabs_to_add
                    self.config_tabs[plugin_name] = plugin_tabs

        elif (hasattr(plugin, 'CONFIGWIDGET_CLASS') and
                plugin.CONFIGWIDGET_CLASS is not None):
            # Old API
            Widget = plugin.CONFIGWIDGET_CLASS

            self.config_pages[plugin.CONF_SECTION] = (
                self.OLD_API, Widget, plugin)

    def deregister_plugin_preferences(
            self, plugin: Union[SpyderPluginV2, SpyderPlugin]):
        """Remove a plugin preference page and additional configuration tabs."""
        name = (getattr(plugin, 'NAME', None) or
                    getattr(plugin, 'CONF_SECTION', None))

        # Remove configuration page for the plugin
        self.config_pages.pop(name)

        # Remove additional configuration tabs that the plugin did introduce
        if isinstance(plugin, SpyderPluginV2):
            for plugin_name in (plugin.ADDITIONAL_CONF_TABS or []):
                tabs = plugin.ADDITIONAL_CONF_TABS[plugin_name]
                for tab in tabs:
                    self.config_tabs[plugin_name].remove(tab)

    def check_version_and_merge(self, conf_section: str, conf_key: str,
                                new_value: BasicType,
                                current_version: Version, plugin):
        """Add a versioned additional option to a configuration section."""
        current_value = self.get_conf(
            conf_key, section=conf_section, default=None)
        section_additional = self.get_conf('additional_configuration',
                                           section=conf_section,
                                           default={})
        plugin_additional = section_additional.get(plugin.NAME, {})

        if conf_key in plugin_additional:
            conf_key_info = plugin_additional[conf_key]
            prev_default = conf_key_info['default']
            prev_version = parse_version(conf_key_info['version'])

            allow_replacement = current_version > prev_version
            allow_deletions = current_version.major > prev_version.major
            new_value = self.merge_defaults(prev_default, new_value,
                                            allow_replacement, allow_deletions)
            new_default = new_value

            if current_value != NoDefault:
                new_value = self.merge_configurations(current_value, new_value)

            self.set_conf(
                conf_key, new_value, section=conf_section)

            conf_key_info['version'] = str(current_version)
            conf_key_info['default'] = new_default
            plugin_additional[conf_key] = conf_key_info
            section_additional[plugin.NAME] = plugin_additional

            self.set_conf(
                'additional_configuration', section_additional,
                section=conf_section)
        else:
            plugin_additional[conf_key] = {
                'version': str(current_version),
                'default': new_value
            }
            section_additional[plugin.NAME] = plugin_additional

            self.set_conf(
                'additional_configuration', section_additional,
                section=conf_section)

            if current_value != NoDefault:
                new_value = self.merge_configurations(current_value, new_value)

            self.set_conf(
                conf_key, new_value, section=conf_section)


    def merge_defaults(self, prev_default: BasicType,
                       new_default: BasicType,
                       allow_replacement: bool = False,
                       allow_deletions: bool = False) -> BasicType:
        """Compare and merge two versioned values."""
        prev_type = type(prev_default)
        new_type = type(new_default)

        if prev_type is dict and new_type is dict:
            # Merge two dicts case
            for new_key in new_default:
                if new_key in prev_default:
                    current_subvalue = prev_default[new_key]
                    new_subvalue = new_default[new_key]
                    prev_default[new_key] = self.merge_defaults(
                        current_subvalue, new_subvalue,
                        allow_replacement, allow_deletions)
                else:
                    # Additions are allowed everytime
                    prev_default[new_key] = new_default[new_key]

            if allow_deletions:
                for old_key in list(prev_default.keys()):
                    if old_key not in new_default:
                        prev_default.pop(old_key)

            return prev_default
        elif prev_default != new_default:
            if allow_replacement:
                return new_default
            else:
                return prev_default
        else:
            return prev_default

    def merge_configurations(
            self, current_value: BasicType, new_value: BasicType) -> BasicType:
        """
        Recursively match and merge a new configuration value into a
        previous one.
        """
        current_type = type(current_value)
        new_type = type(new_value)
        iterable_types = {list, tuple}
        base_types = {int, float, bool, complex, str, bytes}

        if current_type is dict and new_type is dict:
            # Merge two dicts case
            for new_key in new_value:
                if new_key in current_value:
                    current_subvalue = current_value[new_key]
                    new_subvalue = new_value[new_key]
                    current_value[new_key] = self.merge_configurations(
                        current_subvalue, new_subvalue)
                else:
                    current_value[new_key] = new_value[new_key]
            return current_value
        elif current_type in iterable_types and new_type in iterable_types:
            # Merge two lists/tuples case
            return current_type(list(current_value) + list(new_value))
        elif (current_type == new_type and
                current_type in base_types and new_type in base_types):
            # Replace the values directly
            return new_value
        elif current_type in iterable_types and new_type in base_types:
            # Add a value to a list or tuple
            return current_type((list(current_value) + [new_value]))
        elif current_value is None:
            # Assigns the new value if it doesn't exist
            return new_value
        else:
            logger.warning(f'The value {current_value} cannot be replaced'
                           f'by {new_value}')
            return current_value

    def open_dialog(self, prefs_dialog_size):
        container = self.get_container()
        container.create_dialog(
            self.config_pages, self.config_tabs, prefs_dialog_size,
            self.get_main())

    # ---------------- Public Spyder API required methods ---------------------
    @staticmethod
    def get_name() -> str:
        return _('Preferences')

    def get_description(self) -> str:
        return _('This plugin provides access to Spyder preferences page')

    def get_icon(self) -> QIcon:
        return self.create_icon('configure')

    def on_initialize(self):
        container = self.get_container()
        main = self.get_main()

        container.sig_show_preferences_requested.connect(
            lambda: self.open_dialog(main.prefs_dialog_size))
        container.sig_reset_preferences_requested.connect(self.reset)

    @on_plugin_available(plugin=Plugins.MainMenu)
    def on_main_menu_available(self):
        container = self.get_container()
        main_menu = self.get_plugin(Plugins.MainMenu)

        main_menu.add_item_to_application_menu(
            container.show_action,
            menu_id=ApplicationMenus.Tools,
            section=ToolsMenuSections.Tools,
        )

        main_menu.add_item_to_application_menu(
            container.reset_action,
            menu_id=ApplicationMenus.Tools,
            section=ToolsMenuSections.Extras,
        )

    @on_plugin_available(plugin=Plugins.Toolbar)
    def on_toolbar_available(self):
        container = self.get_container()
        toolbar = self.get_plugin(Plugins.Toolbar)
        toolbar.add_item_to_application_toolbar(
            container.show_action,
            toolbar_id=ApplicationToolbars.Main,
            section=MainToolbarSections.ApplicationSection,
            before=PythonpathActions.Manager
        )

    @on_plugin_teardown(plugin=Plugins.MainMenu)
    def on_main_menu_teardown(self):
        main_menu = self.get_plugin(Plugins.MainMenu)

        main_menu.remove_item_from_application_menu(
            PreferencesActions.Show,
            menu_id=ApplicationMenus.Tools,
        )

        main_menu.remove_item_from_application_menu(
            PreferencesActions.Reset,
            menu_id=ApplicationMenus.Tools,
        )

    @on_plugin_teardown(plugin=Plugins.Toolbar)
    def on_toolbar_teardown(self):
        toolbar = self.get_plugin(Plugins.Toolbar)
        toolbar.remove_item_from_application_toolbar(
            PreferencesActions.Show,
            toolbar_id=ApplicationToolbars.Main
        )

    @Slot()
    def reset(self):
        answer = QMessageBox.warning(self.main, _("Warning"),
             _("Spyder will restart and reset to default settings: <br><br>"
               "Do you want to continue?"),
             QMessageBox.Yes | QMessageBox.No)
        if answer == QMessageBox.Yes:
            os.environ['SPYDER_RESET'] = 'True'
            self.sig_restart_requested.emit()

    def on_close(self, cancelable=False):
        container = self.get_container()
        if container.is_preferences_open():
            container.close_preferences()
        return True



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2009- Spyder Project Contributors
#
# Distributed under the terms of the MIT License
# (see spyder/__init__.py for details)
# -----------------------------------------------------------------------------

"""
Class to control a file where stanbdard output can be written.
"""

# Standard library imports.
import codecs
import os
import os.path as osp

# Local imports
from spyder.py3compat import to_text_string
from spyder.utils.encoding import get_coding
from spyder.utils.programs import get_temp_dir


def std_filename(connection_file, extension, std_dir=None):
    """Filename to save kernel output."""
    json_file = osp.basename(connection_file)
    file = json_file.split('.json')[0] + extension
    if std_dir is not None:
        file = osp.join(std_dir, file)
    else:
        try:
            file = osp.join(get_temp_dir(), file)
        except (IOError, OSError):
            file = None
    return file


class StdFile:
    def __init__(self, connection_file, extension=None, std_dir=None):
        if extension is None:
            self.filename = connection_file
        else:
            self.filename = std_filename(connection_file, extension, std_dir)
        self._mtime = 0
        self._cursor = 0
        self._handle = None

    @property
    def handle(self):
        """Get handle to file."""
        if self._handle is None and self.filename is not None:
            # Needed to prevent any error that could appear.
            # See spyder-ide/spyder#6267.
            try:
                self._handle = codecs.open(
                    self.filename, 'w', encoding='utf-8')
            except Exception:
                pass
        return self._handle

    def remove(self):
        """Remove file associated with the client."""
        try:
            # Defer closing the handle until the client
            # is closed because jupyter_client needs it open
            # while it tries to restart the kernel
            if self._handle is not None:
                self._handle.close()
            os.remove(self.filename)
            self._handle = None
        except Exception:
            pass

    def get_contents(self):
        """Get the contents of the std kernel file."""
        try:
            with open(self.filename, 'rb') as f:
                # We need to read the file as bytes to be able to
                # detect its encoding with chardet
                text = f.read()

                # This is needed to avoid showing an empty error message
                # when the kernel takes too much time to start.
                # See spyder-ide/spyder#8581.
                if not text:
                    return ''

                # This is needed since the file could be encoded
                # in something different to utf-8.
                # See spyder-ide/spyder#4191.
                encoding = get_coding(text)
                text = to_text_string(text, encoding)
                return text
        except Exception:
            return None

    def poll_file_change(self):
        """Check if the std kernel file just changed."""
        if self._handle is not None and not self._handle.closed:
            self._handle.flush()
        try:
            mtime = os.stat(self.filename).st_mtime
        except Exception:
            return

        if mtime == self._mtime:
            return
        self._mtime = mtime
        text = self.get_contents()
        if text:
            ret_text = text[self._cursor:]
            self._cursor = len(text)
            return ret_text

    def copy(self):
        """Return a copy."""
        return StdFile(self.filename)



# -*- coding: utf-8 -*-

# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

"""General purpose prefix tree, also known as a trie."""


class TrieNode:
    def __init__(self, key=None, value=None):
        self.children = {}
        self.key = key
        self.value = value

    def __setitem__(self, sequence, value):
        elem = sequence[0]
        if self.key is None:
            self.key = elem

        if len(self.key) > 0:
            sequence = sequence[1:]

        if sequence:
            elem = sequence[0]
            node = self.children.get(elem, None)
            if node is None:
                node = TrieNode()
                self.children[elem] = node
            node[sequence] = value
        else:
            self.value = value

    def __getitem__(self, sequence):
        node = None
        if sequence[0] == self.key:
            sequence = sequence[1:]
            if sequence:
                if sequence[0] in self.children:
                    next_children = self.children[sequence[0]]
                    node = next_children[sequence]
            else:
                node = self
        return node

    def __iter__(self):
        queue = [self]
        while queue != []:
            node = queue.pop(0)
            queue += list(node.children.values())
            if node.value is not None:
                yield node

    def __contains__(self, sequence):
        if len(sequence) + len(self.key) == 0:
            return True

        elem = sequence[0]
        if elem == self.key:
            sequence = sequence[1:]
            if not sequence:
                if self.value is not None:
                    return True
                else:
                    return False
            elem = sequence[0]

        found = elem in self.children
        if found:
            next_children = self.children[elem]
            found = sequence in next_children
        return found


class Trie(TrieNode):
    def __init__(self):
        super().__init__('')
        self.sequences = []

    def __getitem__(self, sequence):
        if sequence:
            elem = sequence[0]
            if elem in self.children:
                node = self.children[elem]
                return node[sequence]
        else:
            return self
        return None

    def __setitem__(self, sequence, value):
        if sequence:
            super().__setitem__(sequence, value)
        else:
            self.value = value
